"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const fs = require("fs-extra");
const os = require("os");
const path = require("path");
const decompress = require("decompress");
const url_1 = require("url");
const decompressUnzip = require('decompress-unzip');
const libsodiumTypes_1 = require("./libsodiumTypes");
const sumoOnlySymbols_1 = require("./sumoOnlySymbols");
const utils = require("./utils");
class TypeGenerator {
    constructor(outputFileOrDir, libsodiumLocalSource) {
        this.outputFileOrDir = outputFileOrDir;
        this.libsodiumLocalSource = libsodiumLocalSource;
        this.libsodiumVersion = '0.7.3';
        this.sourceIsSet = false;
        this.tmpDir = '';
        this.additionalSymbols = libsodiumTypes_1.default.additionalSymbols;
        this.enums = libsodiumTypes_1.default.enums;
        this.genericTypes = libsodiumTypes_1.default.genericTypes;
        this.types = libsodiumTypes_1.default.types;
        this.outputFileOrDir = path.resolve(this.outputFileOrDir);
    }
    setDownloadVersion(version) {
        return __awaiter(this, void 0, void 0, function* () {
            const comparison = utils.compareVersionNumbers(version, this.libsodiumVersion);
            if (isNaN(comparison) || comparison < 0) {
                throw new Error(`Minimum version is ${this.libsodiumVersion}.`);
            }
            this.libsodiumVersion = version;
            return this;
        });
    }
    get externalLibsodiumSource() {
        return `https://github.com/jedisct1/libsodium.js/archive/${this.libsodiumVersion}.zip`;
    }
    downloadLibrary() {
        return __awaiter(this, void 0, void 0, function* () {
            console.info(`Downloading libsodium.js from "${this.externalLibsodiumSource}" ...`);
            try {
                const osTmpDir = os.tmpdir();
                this.tmpDir = yield fs.mkdtemp(path.join(osTmpDir, 'ltg-'));
            }
            catch (err) {
                throw new Error(`Could not create temp dir: ${err.message}`);
            }
            const zipURL = new url_1.URL(this.externalLibsodiumSource);
            const downloadFileName = path.join(this.tmpDir, 'libsodium.zip');
            const file = yield utils.httpsGetFileAsync(zipURL, downloadFileName);
            yield decompress(file, this.tmpDir, { plugins: [decompressUnzip()] });
            const proposedSymbolSource = path.join(this.tmpDir, `libsodium.js-${this.libsodiumVersion}`);
            return proposedSymbolSource;
        });
    }
    getVersion() {
        return this.libsodiumVersion;
    }
    getFunctions() {
        return __awaiter(this, void 0, void 0, function* () {
            const symbolPath = path.join(this.libsodiumLocalSource, 'wrapper', 'symbols');
            const symbolFiles = yield fs.readdir(symbolPath);
            const symbols = yield Promise.all(symbolFiles.map((symbolFile) => __awaiter(this, void 0, void 0, function* () {
                const symbolRaw = yield fs.readFile(path.join(symbolPath, symbolFile), {
                    encoding: 'utf-8'
                });
                try {
                    return JSON.parse(symbolRaw);
                }
                catch (error) {
                    throw new Error(`Could not parse "${path.join(symbolPath, symbolFile)}": ${error.message}`);
                }
            })));
            return symbols
                .concat(this.additionalSymbols)
                .sort((a, b) => (a.name < b.name ? -1 : a.name > b.name ? 1 : 0));
        });
    }
    getConstants() {
        return __awaiter(this, void 0, void 0, function* () {
            const filePath = path.join(this.libsodiumLocalSource, 'wrapper', 'constants.json');
            const constantsRaw = yield fs.readFile(filePath, { encoding: 'utf-8' });
            const constants = JSON.parse(constantsRaw.toString());
            if (this.libsodiumVersion)
                constants.push({
                    name: 'ready',
                    type: 'Promise<void>'
                });
            return constants.sort((a, b) => (a.name < b.name ? -1 : a.name > b.name ? 1 : 0));
        });
    }
    convertType(type) {
        switch (type) {
            case 'uint':
                return 'number';
            case 'buf':
            case 'randombytes_implementation':
                return 'Uint8Array';
            case 'unsized_buf':
            case 'unsized_buf_optional':
                return 'string | Uint8Array | undefined';
            default:
                return type;
        }
    }
    convertReturnType(type) {
        if (type.startsWith('{publicKey: _format_output')) {
            return { binaryType: 'KeyPair', stringType: 'StringKeyPair' };
        }
        if (type.startsWith('_format_output({ciphertext: ciphertext, mac: mac}')) {
            return { binaryType: 'CryptoBox', stringType: 'StringCryptoBox' };
        }
        if (type.startsWith('_format_output({mac: mac, cipher: cipher}')) {
            return { binaryType: 'SecretBox', stringType: 'StringSecretBox' };
        }
        if (type.startsWith('_format_output({sharedRx: sharedRx, sharedTx: sharedTx}')) {
            return { binaryType: 'CryptoKX', stringType: 'StringCryptoKX' };
        }
        if (type === 'random_value') {
            return 'number';
        }
        if (type.includes('=== 0')) {
            return 'boolean';
        }
        if (type.includes('stringify')) {
            return 'string';
        }
        if (type.includes('_format_output')) {
            return { binaryType: 'Uint8Array', stringType: 'string' };
        }
        return type;
    }
    buildData(sumo) {
        return __awaiter(this, void 0, void 0, function* () {
            const getParameters = (parameterArr, formattingAvailable) => {
                let parameters = '';
                parameterArr.forEach((param, index) => {
                    const isLast = index === parameterArr.length - 1;
                    const convertedType = this.convertType(param.type);
                    const optional = param.type.includes('optional');
                    parameters +=
                        `${param.name}: ` +
                            convertedType +
                            (optional ? ' | null' : '') +
                            (isLast && !formattingAvailable ? '' : ', ');
                });
                return parameters;
            };
            let data = `// Type definitions for libsodium-wrappers${sumo ? '-sumo' : ''} ${this.libsodiumVersion}\n` +
                '// Project: https://github.com/jedisct1/libsodium.js\n' +
                '// Definitions by: Florian Keller <https://github.com/ffflorian>\n\n' +
                `declare module 'libsodium-wrappers${sumo ? '-sumo' : ''}' {\n`;
            Object.keys(this.types).forEach(typeName => {
                data += `  type ${typeName} = `;
                this.types[typeName].forEach((typeValue, index) => {
                    const isLast = index === this.types[typeName].length - 1;
                    data += typeValue + (isLast ? ';' : ' | ');
                });
                data += '\n';
            });
            data += '\n';
            Object.keys(this.enums).forEach(enumName => {
                data += `  enum ${enumName} {\n`;
                this.enums[enumName].forEach(enumValue => (data += `    ${enumValue},\n`));
                data += `  }\n\n`;
            });
            Object.keys(this.genericTypes).forEach(typeName => {
                data += `  interface ${typeName} {\n`;
                this.genericTypes[typeName].forEach(({ name, type }) => {
                    data += `    ${name}: ${type};\n`;
                });
                data += `  }\n\n`;
            });
            let constants = yield this.getConstants();
            if (!sumo) {
                constants = constants.filter(constant => !sumoOnlySymbols_1.default.includes(constant.name.toLowerCase()));
            }
            constants.forEach(constant => {
                const convertedType = this.convertType(constant.type);
                data += `  const ${constant.name}: ${convertedType};\n`;
            });
            data += '\n';
            let functions = yield this.getFunctions();
            if (!sumo) {
                functions = functions.filter(func => !sumoOnlySymbols_1.default.includes(func.name.toLowerCase()));
            }
            functions.forEach(fn => {
                if (!fn.return) {
                    fn.return = 'void';
                }
                const formattingAvailable = fn.return.includes('_format_output');
                const inputs = fn.inputs
                    ? getParameters(fn.inputs, formattingAvailable)
                    : '';
                const returnType = this.convertReturnType(fn.return);
                if (typeof returnType === 'object') {
                    data +=
                        `  function ${fn.name}` +
                            `(${inputs}outputFormat?: Uint8ArrayOutputFormat | null): ` +
                            returnType.binaryType +
                            ';\n' +
                            `  function ${fn.name}` +
                            `(${inputs}outputFormat: StringOutputFormat | null): ` +
                            returnType.stringType +
                            ';\n';
                }
                else {
                    data += `  function ${fn.name}(${inputs}): ${returnType};\n`;
                }
            });
            return data + '}';
        });
    }
    generate(sumo) {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.libsodiumLocalSource) {
                this.sourceIsSet = true;
            }
            else {
                this.libsodiumLocalSource = yield this.downloadLibrary();
            }
            const version = yield utils.checkSource(this.libsodiumLocalSource);
            this.setDownloadVersion(version);
            const data = yield this.buildData(sumo);
            const outputFileOrDirStats = yield fs.lstat(this.outputFileOrDir);
            if (outputFileOrDirStats.isDirectory()) {
                const fileName = `libsodium-wrappers${sumo ? '-sumo' : ''}.d.ts`;
                this.outputFileOrDir = path.resolve(this.outputFileOrDir, fileName);
            }
            yield fs.writeFile(this.outputFileOrDir, data);
            if (!this.sourceIsSet) {
                yield fs.remove(this.libsodiumLocalSource);
            }
            if (this.tmpDir) {
                yield fs.remove(this.tmpDir);
                this.tmpDir = '';
            }
            return this.outputFileOrDir;
        });
    }
    generateStandard() {
        return __awaiter(this, void 0, void 0, function* () {
            return this.generate(false);
        });
    }
    generateSumo() {
        return __awaiter(this, void 0, void 0, function* () {
            return this.generate(true);
        });
    }
}
exports.default = TypeGenerator;
